/*  Copyright 2010, Unpublished Work of Technologic Systems
 *  All Rights Reserved.
 *
 *  THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 *  PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 *  ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS EMPLOYEES
 *  WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR
 *  ASSIGNMENTS  AND (II) ENTITIES OTHER THAN TECHNOLOGIC SYSTEMS WHO
 *  HAVE ENTERED INTO  APPROPRIATE LICENSE AGREEMENTS.  NO PART OF THIS
 *  WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, DISTRIBUTED, REVISED,
 *  MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, EXPANDED, COLLECTED,
 *  COMPILED,LINKED,RECAST, TRANSFORMED, ADAPTED IN ANY FORM OR BY ANY
 *  MEANS,MANUAL, MECHANICAL, CHEMICAL, ELECTRICAL, ELECTRONIC, OPTICAL,
 *  BIOLOGICAL, OR OTHERWISE WITHOUT THE PRIOR WRITTEN PERMISSION AND
 *  CONSENT OF TECHNOLOGIC SYSTEMS . ANY USE OR EXPLOITATION OF THIS WORK
 *  WITHOUT THE PRIOR WRITTEN CONSENT OF TECHNOLOGIC SYSTEMS  COULD
 *  SUBJECT THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
 */

#include <assert.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <linux/unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <time.h>
#include <limits.h>
#include <sched.h>
#include <ctype.h>
#include <signal.h>

static void poke8(unsigned int, unsigned char);
static void poke16(unsigned int, unsigned short);
static void poke32(unsigned int, unsigned int);
static unsigned char peek8(unsigned int);
static unsigned short peek16(unsigned int);
static unsigned int peek32(unsigned int);

static volatile unsigned int *cvspiregs, *cvgpioregs;

int alarmed = 0;
static void alarmsig(int x) {
	alarmed = 1;
}

static void poke8(unsigned int adr, unsigned char dat) {
	if (!cvspiregs) *(unsigned char *)adr = dat;
	else {
		if (adr & 0x1) {
			cvgpioregs[0] = (0x2<<15|1<<17);
			poke16(adr, dat << 8);
		} else {
			cvgpioregs[0] = (0x2<<15|1<<3);
			poke16(adr, dat);
		}
		cvgpioregs[0] = (0x2<<15|1<<17|1<<3);
	}

}


static void poke16(unsigned int adr, unsigned short dat) {
	unsigned int dummy;
	unsigned int d = dat;
	volatile unsigned int *spi = cvspiregs;

	if (!cvspiregs) {
		*(unsigned short *)adr = dat;
	} else asm volatile (
		"mov %0, %1, lsl #18\n"
		"orr %0, %0, #0x800000\n"
		"orr %0, %0, %2, lsl #3\n"
		"3: ldr r1, [%3, #0x64]\n"
		"cmp r1, #0x0\n"
		"bne 3b\n"
		"2: str %0, [%3, #0x50]\n"
		"1: ldr r1, [%3, #0x64]\n"
		"cmp r1, #0x0\n"
		"beq 1b\n"
		"ldr %0, [%3, #0x58]\n"
		"ands r1, %0, #0x1\n"
		"moveq %0, #0x0\n"
		"beq 3b\n"
		: "+r"(dummy) : "r"(adr), "r"(d), "r"(cvspiregs) : "r1","cc"
	);
}

static void poke32(unsigned int adr, unsigned int dat) {
	if (!cvspiregs) *(unsigned int *)adr = dat;
	else {
		poke16(adr, dat & 0xffff);
		poke16(adr + 2, dat >> 16);
	}
}


static unsigned char peek8(unsigned int adr) {
	unsigned char ret;
	if (!cvspiregs) ret = *(unsigned char *)adr;
	else {
		unsigned short x;
		x = peek16(adr);
		if (adr & 0x1) ret = x >> 8;
		else ret = x & 0xff;
	}
	return ret;
}


static unsigned short peek16(unsigned int adr) {
	unsigned short ret;
	volatile unsigned int *spi = cvspiregs;

	if (!cvspiregs) ret = *(unsigned short *)adr;
	else asm volatile (
		"mov %0, %1, lsl #18\n"
		"2: str %0, [%2, #0x50]\n"
		"1: ldr r1, [%2, #0x64]\n"
		"cmp r1, #0x0\n"
		"beq 1b\n"
		"ldr %0, [%2, #0x58]\n"
		"ands r1, %0, #0x10000\n"
		"bicne %0, %0, #0xff0000\n"
		"moveq %0, #0x0\n"
		"beq 2b\n" 
		: "+r"(ret) : "r"(adr), "r"(cvspiregs) : "r1", "cc"
	);

	return ret;

}


static unsigned int peek32(unsigned int adr) {
	unsigned int ret;
	if (!cvspiregs) ret = *(unsigned int *)adr;
	else {
		unsigned short l, h;
		l = peek16(adr);
		h = peek16(adr + 2);
		ret = (l|(h<<16));
	}
	return ret;
}


static void reservemem(void) {
	char dummy[32768];
	int i, pgsize;
	FILE *maps;

	pgsize = getpagesize();
	mlockall(MCL_CURRENT|MCL_FUTURE);
	for (i = 0; i < sizeof(dummy); i += 4096) {
		dummy[i] = 0;
	}

	maps = fopen("/proc/self/maps", "r"); 
	if (maps == NULL) {
		perror("/proc/self/maps");
		exit(1);
	}
	while (!feof(maps)) {
		unsigned int s, e, i;
		char m[PATH_MAX + 1];
		char perm[16];
		int r = fscanf(maps, "%x-%x %s %*x %*x:%*x %*d",
		  &s, &e, perm);
		if (r == EOF) break;
		assert (r == 3);

		i = 0;
		while ((r = fgetc(maps)) != '\n') {
			if (r == EOF) break;
			m[i++] = r;
		}
		m[i] = '\0';
		assert(s <= e && (s & 0xfff) == 0);
		if (perm[0] == 'r') while (s < e) {
			unsigned char *ptr = (unsigned char *)s;
			*ptr;
			s += pgsize;
		}
	}
}


static int semid = -1;
static int sbuslocked = 0;
static int canadr;
static void sbuslock(void) {
	int r;
	struct sembuf sop;
	if (semid == -1) {
		key_t semkey;
		reservemem();
		semkey = 0x75000000;
		semid = semget(semkey, 1, IPC_CREAT|IPC_EXCL|0777);
		if (semid != -1) {
			sop.sem_num = 0;
			sop.sem_op = 1;
			sop.sem_flg = 0;
			r = semop(semid, &sop, 1);
			assert (r != -1);
		} else semid = semget(semkey, 1, 0777);
		assert (semid != -1);
	}
	sop.sem_num = 0;
	sop.sem_op = -1;
	sop.sem_flg = SEM_UNDO;
	r = semop(semid, &sop, 1);
	assert (r == 0);
	cvgpioregs[0] = (2<<15|1<<17|1<<3);
	sbuslocked = 1;
	canadr = -1;
}


static void sbusunlock(void) {
	struct sembuf sop = { 0, 1, SEM_UNDO};
	int r;
	if (!sbuslocked) return;
	r = semop(semid, &sop, 1);
	assert (r == 0);
	sbuslocked = 0;
}


static void sbuspreempt(void) {
	int r;
	r = semctl(semid, 0, GETNCNT);
	assert (r != -1);
	if (r) {
		sbusunlock();
		sched_yield();
		sbuslock();
	}
}


static unsigned int can_peek8(int adr) {
	if (adr != canadr) poke16(0x10, adr);
	canadr = adr + 1;
	
	return peek16(0x12) & 0xff;
}


static void can_poke8(int adr, unsigned int dat) {
	dat &= 0xff;
	poke16(0x16, adr<<8|dat);
	canadr = adr + 1;
}


static void can_peekstream(int adr, unsigned char *buf, int n) {
	assert(n != 0);

	if (adr != canadr) poke16(0x10, adr);
	canadr = adr + n + 1;

	while (n >= 2) {
		unsigned int s = peek16(0x14);
		*buf++ = s & 0xff;
		*buf++ = s >> 8;
		n -= 2;
	}

	if (n) *buf = peek16(0x12) & 0xff;
};


static void can_pokestream(int adr, unsigned char *buf, int n) {
	int na = adr + n + 1;
	assert(n != 0);

	if (adr != canadr) {
		can_poke8(adr, *buf++);
		n--;
	} 
	
	canadr = na;

	while (n >= 2) {
		unsigned int s;
		s = *buf++;
		s |= *buf++ << 8;
		poke16(0x14, s);
		n -= 2;
	}

	if (n) poke16(0x12, *buf);
}

#define CAN_RTR	0x10
static void can_loadtx(unsigned int id, unsigned char *d, int flags) {
	unsigned char buf[13];
	unsigned char *b = buf;
	int dlc = flags & 0xf;
	int i;

	assert(dlc <= 8);

	if (flags & CAN_RTR) *b = 0xc0; else *b = 0x80;
	*b++ |= dlc;
	id <<= 3;
	*b++ = (id >> 24);
	*b++ = (id >> 16);
	*b++ = (id >> 8);
	*b++ = (id & 0xff);
	while (dlc--) *b++ = *d++;

	can_pokestream(16, buf, b - buf);
}


static void can_rxpacket(unsigned char *d) {
	unsigned int t, len;
	int i;
	if (canadr != 16) poke16(0x10, 16);

	t = peek16(0x14);
	d[0] = t & 0xff;
	d[1] = t >> 8;
	canadr = 18;

	/* standard or extended frame */
	if (d[0] & 0x80) len = 3;
	else len = 1;

	if (!(d[0] & 0x40)) /* Not RTR packet? */
	  len += ((d[0] & 0xf) > 8) ? 8 : (d[0] & 0xf);

	can_peekstream(18, &d[2], len);
}


static int can_id(unsigned char *d) {
	int ff = d[0] & 0x80;
	int id;

	if (ff) {
		id = d[1] << 21;
		id |= d[2] << 13;
		id |= d[3] << 5;
		id |= d[4] >> 3;
	} else {
		id = d[1] << 21;
		id |= (d[2] & 0xe0) << 13;
		id = id >> 18;
	}

	return id;
}


static void dump_packet(unsigned char *d) {
	struct timeval tv;
	struct tm *tm;
	int ff = d[0] & 0x80;
	int rtr = d[0] & 0x40;
	int id, i;
	int dlc = d[0] & 0xf;

	if (ff) {
		id = d[1] << 21;
		id |= d[2] << 13;
		id |= d[3] << 5;
		id |= d[4] >> 3;
		d = &d[5];
	} else {
		id = d[1] << 21;
		id |= (d[2] & 0xe0) << 13;
		id = id >> 18;
		d = &d[3];
	}

	gettimeofday(&tv, NULL);
	tm = localtime(&tv.tv_sec);

	printf("%02d:%02d:%02d.%06d %c%c ", tm->tm_hour, tm->tm_min, tm->tm_sec,
	  tv.tv_usec, ff ? 'E' : 'S', rtr ? 'R' : '.');
	
	printf("id=0x%08x ", id);

	if (!rtr && dlc) {
		printf("dat=");
		if (dlc > 8) dlc = 8;
		for (i = 0; i < dlc; i++) {
			printf("0x%02x", *d++);
			if (i != (dlc - 1)) printf(":");
		}
	}
	printf("\n");
}


static int parse_txdat(char *s, unsigned char *txdat) {
	unsigned int val, len;
	char *e;

	len = 0;
	while (*s) {
		val = strtoul(s, &e, 0);
		if (s == e) return -1;
		else if (ispunct(*e)) s = e + 1; 
		else s = e;
		*txdat++ = val;
		len++;
		if (len == 8) break;
	}

	return len;	
}


static void can_irqwait(void) {
	static int can_irqfd;
	static int inited;
	int dummy;

	if (!inited) {
		can_irqfd = open("/proc/irq/30/irq", O_RDONLY|O_SYNC);
		assert(can_irqfd != -1);
		inited = 1;
	}

	read(can_irqfd, &dummy, sizeof(dummy));
}


void usage(char **argv) {
	fprintf(stderr, "Usage: %s [OPTION] ...\n"
	  "Technologic Systems CAN controller manipulation.\n"
	  "\n"
	  "General options:\n"
	  "  -a, --address=ADR       CAN register address\n"
	  "  -r, --peek8             CAN register read\n"
	  "  -w, --poke8=VAL         CAN register write\n"
	  "  -t, --txid=ID           CAN TX packet ID\n"
	  "  -d, --txdat=DAT         TX packet with data DAT\n"
	  "  -R, --txrtr             TX RTR packet\n"
	  "  -D, --dump              Receive and print all CAN packets\n"
	  "  -0, --btr0=BTR0         SJA1000 BTR0 bus timing reg val\n"
	  "  -1, --btr1=BTR1         SJA1000 BTR1 bus timing reg val\n"
	  "  -x, --txtest            Send TX test pattern\n"
	  "  -y, --rxtest            Do RX test\n"
	  "  -h, --help              This help\n"
	  "\n",
	  argv[0]
	);
}


int main(int argc, char **argv) {
	int devmem, i, j, c, k, ret, txdat_len;
	int opt_address = 0, opt_pokeval = 0, opt_peek8 = 0, opt_poke8 = 0;
	int opt_txrtr = 0, opt_txdat = 0, opt_dump = 0, opt_rxtest = 0;
	int opt_btr0 = 0, opt_btr1 = 0x7e, opt_txtest = 0;
	unsigned int opt_txid = 0, ui;
	unsigned char txdat[8];
	static struct option long_options[] = {
	  { "address", 1, 0, 'a' },
	  { "peek8", 0, 0, 'r' },
	  { "poke8", 1, 0, 'w' },
	  { "txid", 1, 0, 't' },
	  { "txrtr", 0, 0, 'R' },
	  { "txdat", 1, 0, 'd' },
	  { "dump", 0, 0, 'D' },
	  { "btr0", 1, 0, '0' },
	  { "btr1", 1, 0, '1' },
	  { "txtest", 0, 0, 'x' },
	  { "rxtest", 0, 0, 'y' },
	  { "help", 0, 0, 'h' },
	  { 0, 0, 0, 0 }
	};


	while((c = getopt_long(argc, argv,
	  "a:w:rt:d:RD1:0:xyh", long_options, NULL)) != -1) {
		switch (c) {
		case 'y':
			opt_rxtest = 1;
			break;
		case 'x':
			opt_txtest = 1;
			break;
		case '0':
			opt_btr0 = strtoul(optarg, NULL, 0);
			break;
		case '1':
			opt_btr1 = strtoul(optarg, NULL, 0);
			break;
		case 'D':
			opt_dump = 1;
			break;
		case 'a':
			opt_address = strtoul(optarg, NULL, 0);
			break;
		case 'w':
			opt_pokeval = strtoul(optarg, NULL, 0);
			opt_poke8 = 1;
			break;
		case 'r':
			opt_peek8 = 1;
			break;
		case 't':
			opt_txid = strtoul(optarg, NULL, 0);
			break;
		case 'R':
			opt_txrtr = 1;
			break;
		case 'd':
			opt_txdat = 1;
			txdat_len = parse_txdat(optarg, txdat);
			if (txdat_len == -1) {
				fprintf(stderr, "txdat parse error\n");
				return(1);
			};
			break;
		case 'h':
		default:
			usage(argv);
			return(1);
		}
	} 

	
	devmem = open("/dev/mem", O_RDWR|O_SYNC);
	assert(devmem != -1);
	cvspiregs = (unsigned int *) mmap(0, 4096,
	  PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0x71000000);
	cvgpioregs = (unsigned int *) mmap(0, 4096,
	  PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0x7c000000);

	sbuslock();
	cvspiregs[0x64 / 4] = 0x0; /* RX IRQ threahold 0 */
	cvspiregs[0x40 / 4] = 0x80000c02; /* 24-bit mode, no byte swap */
	cvspiregs[0x60 / 4] = 0x0; /* 0 clock inter-transfer delay */
	cvspiregs[0x6c / 4] = 0x0; /* disable interrupts */
	cvspiregs[0x4c / 4] = 0x4; /* deassert CS# */
	for (i = 0; i < 8; i++) cvspiregs[0x58 / 4];
	cvgpioregs[0] = (2<<15|1<<17|1<<3);
	sbusunlock();

	if (opt_poke8) {
		sbuslock();
		can_poke8(opt_address, opt_pokeval);
		sbusunlock();
		return 0;
	}

	if (opt_peek8) {
		unsigned int x;
		sbuslock();
		x = can_peek8(opt_address);
		sbusunlock();
		printf("0x%x\n", x);
		return 0;
	}

	sbuslock();
	can_poke8(0, 0x1); /* Enter reset mode */
	can_poke8(31, 0x80); /* Enable PeliCAN mode */
	can_poke8(4, 0xff); /* All interrupts enabled */
	can_poke8(6, opt_btr0); /* BTR0 for 500kbps */
	can_poke8(7, opt_btr1); /* BTR1 for 500kbps */
	can_poke8(13, 96); /* Error warning limit */
	can_poke8(14, 0); /* Reset RX error counter to 0 */
	can_poke8(15, 0); /* Reset TX error counter to 0 */

	can_poke8(20, 0xff); /* Acceptance mask disabled (all packets rx'ed) */
	can_poke8(21, 0xff);
	can_poke8(22, 0xff);
	can_poke8(23, 0xff);

	can_peek8(3); /* Clear pending interrupts */

	can_poke8(0, 0); /* Enter operational mode */
	sbusunlock();

	if (opt_txdat || opt_txrtr) {
		sbuslock();
		can_loadtx(opt_txid, txdat, opt_txrtr ? CAN_RTR : txdat_len);
		can_poke8(1, 0x1); /* TX command */
		do {
			sbusunlock();
			can_irqwait();
			sbuslock();
			j = can_peek8(2); /* Poll status */
			can_peek8(3); /* Clear pending interrupts */
		} while ((j & 0x44) == 0);
		if ((j & 0x40) && can_peek8(15) >= 96) {
			can_poke8(1, 0x2); /* Abort transmission */
			j = can_peek8(12); /* error capture reg */
			can_poke8(0, 1); /* Enter reset mode */
			sbusunlock();
			printf("txerror=0x%x\n", j);
			return 1;
		}
		can_poke8(0, 1); /* Enter reset mode */
		sbusunlock();
		return 0;
	} else if (opt_dump) for(;;) { 
		unsigned char tmp[16];
		int err = 0;

		can_irqwait();
		sbuslock();
		j = can_peek8(3); /* read IRQ */
		if (j & 0x1) {
			can_rxpacket(tmp);
			can_poke8(1, 0x4); /* Release receive buffer */
		}
		if (j & 0x8) can_poke8(1, 0x8); /* clear data overrun */
		sbusunlock();

		if (j & 0x80) printf("*** bus error\n");
		if (j & 0x20) printf("*** error passive\n");
		if (j & 0x8) printf("*** data overrun error\n");
		if (j & 0x4) printf("*** error warning\n");
		if (j & 0x1) dump_packet(tmp);
	} else if (opt_txtest) for(sbuslock(), ui = 0; ; ui++) {
		for (j = 0; j < 8; j++) txdat[j] = (ui << 3) + j + 1;
		can_loadtx(ui, txdat, 8);
		can_poke8(1, 0x1); /* TX command */
		do {
			sbusunlock();
			can_irqwait();
			sbuslock();
			j = can_peek8(3); /* pending interrupts */
		} while ((j & 0x2) == 0); /* wait til TX complete */
	} else if (opt_rxtest) {
		struct sigaction sa;
		struct timeval tv1, tv2;
		unsigned char tmp[16];
		int idseq = -1, err = 1, id;
		int corrupts = 0;
		struct sched_param sched;
		struct rusage ru1, ru2;

		nice(-20);
		sched.sched_priority = 50;
		sched_setscheduler(0, SCHED_FIFO, &sched);

		sa.sa_handler = alarmsig;
		sigemptyset(&sa.sa_mask);
		sa.sa_flags = 0;
		sigaction(SIGALRM, &sa, NULL);
		alarm(10);
		gettimeofday(&tv1, NULL);
		getrusage(RUSAGE_SELF, &ru1);
		i = 0;
		while (!alarmed) {
			can_irqwait();
			sbuslock();
			j = can_peek8(3); /* read IRQ */
			if (j & 0x80) 
			  printf("errormsg=\"bus error irq\"\n");
			if (j & 0x20) 
			  printf("errormsg=\"error passive irq\"\n");
			if (j & 0x8) 
			  printf("errormsg=\"data overrun irq\"\n");
			if (j & 0x4) 
			  printf("errormsg=\"error warning irq\"\n");
			if (j & 0x1) {
				i++;
				can_rxpacket(tmp);
				id = can_id(tmp);
				if (idseq == -1) {
					err = 0;
					idseq = id;
				} else {
					int nid = (idseq + 1) & 0x1fffffff;
					if (id != nid) {
						sbusunlock();
						fprintf(stderr, 
						  "errormsg=\""
						  "ID 0x%x != 0x%x\"\n",
						  id, nid);
						sbuslock();
						err++;
					}
					for (k = 0; k < (tmp[0] & 0xf); k++) 
					  if (tmp[5+k] != (((id<<3)+k+1)&0xff)) {
						sbusunlock();
						fprintf(stderr, 
						  "errormsg=\""
						  "data[%d] 0x%x != 0x%x\"\n", 
						  k, tmp[5+k], 
						  (((id<<3)+k+1)&0xff));
						sbuslock();
						corrupts++;
					}
					idseq = id;
				}
				can_poke8(1, 0x4); /* release RX buf */
			}
			sbusunlock();
		}
		/* count packets in i */
		getrusage(RUSAGE_SELF, &ru2);
		gettimeofday(&tv2, NULL);
		j = (tv2.tv_sec - tv1.tv_sec) * 1000;
		j += tv2.tv_usec / 1000 - tv1.tv_usec / 1000;
		printf("packets_per_sec=%d\n", (i * 1000) / j);
		i = (ru2.ru_utime.tv_sec + ru2.ru_stime.tv_sec) * 1000;
		i += ru2.ru_utime.tv_usec / 1000 + ru2.ru_stime.tv_usec / 1000;
		i -= (ru1.ru_utime.tv_sec + ru1.ru_utime.tv_sec) * 1000;
		i -= ru1.ru_utime.tv_usec / 1000 + ru1.ru_stime.tv_usec / 1000;
		printf("cpu_utilization_pct=%d\n", i * 100 / j);
		printf("test_sequence_errors=%d\n", err);
		printf("test_corruption_errors=%d\n", corrupts);
	}

	return 0;
}
